# -*- encoding: ascii -*-
"""
Doc Tasks
~~~~~~~~~

"""

import os as _os

import invoke as _invoke

from . import compile as _compile


@_invoke.task(_compile.compile)
def userdoc(ctx):
    """ Create userdocs """
    _sphinx(ctx, ctx.doc.sphinx.build, ctx.doc.sphinx.source, ctx.doc.userdoc)


@_invoke.task(userdoc, default=True)
def doc(ctx):  # pylint: disable = unused-argument
    """ Create docs """


@_invoke.task()
def website(ctx):
    """ Create website """
    ctx.shell.rm_rf(ctx.doc.website.source)
    ctx.shell.cp_r(ctx.doc.sphinx.source,
                   _os.path.join(ctx.doc.website.source, 'src'))
    ctx.shell.rm_rf(ctx.doc.website.target)
    ctx.shell.mkdir_p(ctx.doc.website.target)

    dlfile = _os.path.join(ctx.doc.website.source, 'src',
                           'website_download.txt')
    with open(dlfile, 'rb') as fp:
        download = fp.read().decode('latin-1')

    ixfile = _os.path.join(ctx.doc.website.source, 'src', 'index.txt')
    with open(ixfile, 'rb') as fp:
        index = fp.read().decode('latin-1').splitlines(True)

    with open(ixfile, 'wb') as fp:
        for line in index:
            if line.startswith('.. placeholder: Download'):
                line = download
            fp.write(line.encode('latin-1'))

    _sphinx(ctx,
            ctx.shell.native(_os.path.join(ctx.doc.website.source, 'build')),
            ctx.shell.native(_os.path.join(ctx.doc.website.source, 'src')),
            ctx.shell.native(ctx.doc.website.target))
    ctx.shell.rm(_os.path.join(ctx.doc.website.target, '.buildinfo'))


def _sphinx(ctx, build, source, target):
    """ Run sphinx """
    sphinx = ctx.shell.frompath('sphinx-build')
    if sphinx is None:
        raise RuntimeError("sphinx-build not found")

    with ctx.shell.root_dir():
        ctx.run(ctx.c(r'''
            %(sphinx)s -a -d %(doctrees)s -b html %(source)s %(target)s
        ''', **dict(
            sphinx=sphinx,
            doctrees=_os.path.join(build, 'doctrees'),
            source=source,
            target=target,
        )), echo=True)



# -*- coding: ascii -*-
#
# Copyright 2007 - 2021
# Andr\xe9 Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
=================
 Shell utilities
=================

Shell utilities.
"""
from __future__ import absolute_import

__author__ = "Andr\xe9 Malo"
__docformat__ = "restructuredtext en"

import contextlib as _contextlib
import errno as _errno
import fnmatch as _fnmatch
import functools as _ft
import os as _os
import re as _re
import shutil as _shutil
import sys as _sys
import tempfile as _tempfile

# pylint: disable = invalid-name

root = _os.path.dirname(_os.path.dirname(_os.path.abspath(__file__)))


@_contextlib.contextmanager
def root_dir():
    """ Context manager to change into the root directory """
    assert root is not None

    old = _os.getcwd()
    try:
        _os.chdir(root)
        yield root
    finally:
        _os.chdir(old)


def _make_split_command():
    """
    Make split_command function

    The command splitter splits between tokens. Tokens are non-whitespace
    sequences or double quoted strings. Inside those double quotes can be
    escaped with a backslash. So have to be backslashes.

    Stolen from <http://opensource.perlig.de/svnmailer/>.

    :Return: Parser for generic commandlines
    :Rtype: callable
    """
    argre = r'[^"\s]\S*|"[^\\"]*(?:\\[\\"][^\\"]*)*"'
    check = _re.compile(
        r'\s*(?:%(arg)s)(?:\s+(?:%(arg)s))*\s*$' % dict(arg=argre)
    ).match
    split = _re.compile(argre).findall
    strip = _ft.partial(_re.compile(r'\\([\\"])').sub, r'\1')

    def split_command(command):  # pylint: disable = redefined-outer-name
        """
        Split generic commandline into single arguments

        The command splitter splits between tokens. Tokens are non-whitespace
        sequences or double quoted strings. Inside those double quotes can be
        escaped with a backslash. So have to be backslashes.

        Stolen from <http://opensource.perlig.de/svnmailer/>.

        :Return: Parser for generic commandlines
        :Rtype: callable
        """
        if not check(command):
            raise ValueError("Invalid command string %r" % (command,))

        return [
            strip(arg[1:-1]) if arg.startswith('"') else arg
            for arg in split(command)
        ]

    return split_command

split_command = _make_split_command()


def _make_formatter(*args, **kwargs):
    """
    Make args / kwargs formatter

    Either args or kwargs or neither of them can be set. There cannot be set
    both of them.

    :Return: Formatter, using either args or kwargs
    :Rtype: callable
    """
    # pylint: disable = no-else-return

    assert not(args and kwargs)

    if args:
        # tuples are given for the whole command string but applied per token.
        # We need to supply only the tuples which are needed for the current
        # token.
        args = list(args[::-1])
        pcents = _re.compile(r'%[^%]').findall

        def formatter(value):
            """ Tuple formatter """
            count = len(pcents(value))
            torepl = []
            while len(torepl) < count:
                torepl.append(args.pop())
            return value % tuple(torepl)
        return formatter

    elif kwargs:
        return lambda x: x % kwargs

    return lambda x: x


def _make_win32_command():
    r"""
    Make win32_command function

    >>> x = win32_command(r'''
    ...     command arg "arg 2" "" "arg %3"
    ...     "malic'ious argument\\\"&whoami"
    ... ''')
    >>> print(x[:42])
    command arg ^"arg^ 2^" ^"^" ^"arg^ ^%3^" ^
    >>> print(x[41:])
    ^"malic'ious^ argument\\\^"^&whoami^"

    """
    wsp, meta = r'\r\n\t\x0b\x0c\x08 ', r'()%!^"<>&|'
    slashsub = _ft.partial(_re.compile(r'(\\+)("|$)').sub, r'\1\1\2')
    metasub = _ft.partial(_re.compile(r'([%s%s])' % (wsp, meta)).sub, r'^\1')
    qsearch = _re.compile(r'[%s"]' % (wsp,)).search
    needq = lambda x: not x or qsearch(x)

    def win32_command(command, *args, **kwargs):
        """
        Return a win32/cmd.exe suitable commandline

        :See: https://blogs.msdn.microsoft.com/twistylittlepassagesallalike/
              2011/04/23/everyone-quotes-command-line-arguments-the-wrong-way/

        Either args or kwargs or neither of them can be set. There cannot be
        set both of them.

        :Parameters:
          `command` : ``str``
            Generic commandline, possibly containing substitutions, filled by
            args or kwargs. See `split_command` for generic commandline
            syntax.

          `args` : ``tuple``
            Substitution tuple

          `kwargs` : ``dict``
            Substitution dict

        :Return: Strictly quoted shell commandline for ``cmd.exe``
        :Rtype: ``str``
        """
        # pylint: disable = redefined-outer-name
        return ' '.join([metasub(
            '"%s"' % (slashsub(token).replace('"', '\\"'),)
            if needq(token) else token
        ) for token in map(_make_formatter(*args, **kwargs),
                           split_command(command))])

    return win32_command

win32_command = _make_win32_command()


def _make_posix_command():
    r"""
    Make posix_command function

    >>> x = posix_command(r'''
    ...     command arg "arg 2" "" "arg $3"
    ...     "malic'ious argument\\\"&whoami"
    ... ''')
    >>> print(x)
    command arg 'arg 2' '' 'arg $3' 'malic'\''ious argument\"&whoami'

    """
    qsearch = _re.compile(r'[^a-zA-Z\d_./-]').search
    needq = lambda x: not x or qsearch(x)

    def posix_command(command, *args, **kwargs):
        """
        Return a POSIX shell suitable commandline

        Either args or kwargs or neither of them can be set. There cannot be
        set both of them.

        :Parameters:
          `command` : ``str``
            Generic commandline, possibly containing substitutions, filled by
            args or kwargs. See `split_command` for generic commandline
            syntax.

          `args` : ``tuple``
            Substitution tuple

          `kwargs` : ``dict``
            Substitution dict

        :Return: Strictly quoted shell commandline for POSIX shells
        :Rtype: ``str``
        """
        # pylint: disable = redefined-outer-name
        return ' '.join([
            "'%s'" % (token.replace("'", "'\\''")) if needq(token) else token
            for token in map(_make_formatter(*args, **kwargs),
                             split_command(command))
        ])
    return posix_command

posix_command = _make_posix_command()

command = win32_command if _sys.platform.lower() == 'win32' else posix_command


def native(path):
    """
    Convert slash path to native

    :Parameters:
      `path` : ``str``
        Path relative to the checkout root

    :Return: The native path
    :Rtype: ``str``
    """
    path = _os.path.sep.join(path.split('/'))
    return _os.path.normpath(_os.path.join(root, path))


def cp(src, dest):
    """
    Copy src to dest

    :Parameters:
      `src` : ``str``
        Source path, relative to the checkout root

      `dest` : ``str``
        Dest path, relative to the checkout root
    """
    _shutil.copy2(native(src), native(dest))


def cp_r(src, dest, ignore=None):
    """
    Copy -r src to dest

    :Parameters:
      `src` : ``str``
        Source path, relative to the checkout root

      `dest` : ``str``
        Dest path, relative to the checkout root

      `ignore` : callable
        Ignore callback
    """
    _shutil.copytree(native(src), native(dest), ignore=ignore)


def rm(*dest):
    """
    Remove a file, ENOENT is not considered an error

    :Parameters:
      `dest` : ``str``
        File to remove
    """
    for name in dest:
        try:
            _os.unlink(native(name))
        except OSError as e:
            if _errno.ENOENT != e.errno:
                raise


def rm_rf(*dest):
    """
    Remove a tree

    :Parameters:
      `dest` : ``str``
        Path to remove
    """
    for name in dest:
        name = native(name)
        if _os.path.exists(name):
            if _os.path.islink(name):
                _os.unlink(name)
                continue

            for path in files(name, '*'):
                if not _os.path.islink(native(path)):
                    _os.chmod(native(path), 0o644)
            _shutil.rmtree(name)


def mkdir_p(dirname):
    """
    Create direcories

    :Parameters:
      `dirname` : ``str``
        Directory name (the leaf directory)
    """
    try:
        _os.makedirs(dirname)
    except OSError as e:
        # makedirs throws OSError if the last dir segment exists
        if e.errno != _errno.EEXIST:
            raise


mkstemp = _tempfile.mkstemp
walk = _os.walk


def files(base, wildcard='[!.]*', recursive=1, prune=('.git', '.svn', 'CVS')):
    """
    Determine a filelist

    :Parameters:
      `base` : ``str``
        Base path to start from

      `wildcard` : ``str``
        Glob to match against

      `recursive` : ``bool``
        Deep walk into the tree? Default: true

      `prune` : iterable
        List of directory basenames to ignore.
        Default: ('.git', '.svn', 'CVS'). Can be empty or ``None`` (meaning
        the same)

    :Return: Iterator over matching pathnames
    :Rtype: iterable
    """
    prune = tuple(prune or ())
    for dirpath, dirnames, filenames in walk(native(base)):
        for item in prune:
            if item in dirnames:
                dirnames.remove(item)

        filenames.sort()
        for name in _fnmatch.filter(filenames, wildcard):
            dest = _os.path.join(dirpath, name)
            if dest.startswith(root):
                dest = dest.replace(root, '', 1)
            aslist = []
            head, tail = _os.path.split(dest)
            while tail:
                aslist.append(tail)
                head, tail = _os.path.split(head)
            aslist.reverse()
            dest = '/'.join(aslist)
            yield dest

        if not recursive:
            break
        dirnames.sort()


def dirs(base, wildcard='[!.]*', recursive=1, prune=('.git', '.svn', 'CVS')):
    """
    Determine a directory list

    :Parameters:
      `base` : ``str``
        Base path to start from

      `wildcard` : ``str``
        Glob to match against

      `recursive` : ``bool``
        Deep walk into the tree? Default: true

      `prune` : iterable
        List of directory basenames to ignore.
        Default: ('.git', '.svn', 'CVS'). Can be empty or ``None`` (meaning
        the same)

    :Return: Iterator over matching pathnames
    :Rtype: iterable
    """
    prune = tuple(prune or ())
    for dirpath, dirnames, _ in walk(native(base)):
        for item in prune:
            if item in dirnames:
                dirnames.remove(item)

        dirnames.sort()
        for name in _fnmatch.filter(dirnames, wildcard):
            dest = _os.path.join(dirpath, name)
            if dest.startswith(root):
                dest = dest.replace(root, '', 1)
            aslist = []
            head, tail = _os.path.split(dest)
            while tail:
                aslist.append(tail)
                head, tail = _os.path.split(head)
            aslist.reverse()
            dest = '/'.join(aslist)
            yield dest

        if not recursive:
            break


def frompath(executable):
    """
    Find executable in PATH

    :Parameters:
      `executable` : ``str``
        Command to search for

    :Return: Full path or ``None``
    :Rtype: ``str``
    """
    # Based on distutils.spawn.find_executable.
    path = _os.environ.get('PATH', '')
    paths = [
        _os.path.expanduser(item)
        for item in path.split(_os.pathsep)
    ]
    ext = _os.path.splitext(executable)[1]
    exts = ['']
    if _sys.platform == 'win32' or _os.name == 'os2':
        eext = ['.exe', '.bat', '.py']
        if ext not in eext:
            exts.extend(eext)

    for ext in exts:
        if not _os.path.isfile(executable + ext):
            for path in paths:
                fname = _os.path.join(path, executable + ext)
                if _os.path.isfile(fname):
                    # the file exists, we have a shot at spawn working
                    return fname
        else:
            return executable + ext

    return None



# -*- encoding: ascii -*-
"""
Uploading
~~~~~~~~~

"""

import invoke as _invoke


@_invoke.task(default=True)
def source(ctx):
    """ Upload source package """
    with ctx.shell.root_dir():
        files = list(ctx.shell.files('dist', '*.tar.gz'))
        if len(files) != 1:
            ctx.fail("Not exactly one tarball found")

        ctx.run(ctx.c('''
            twine upload
            --repository-url %s
            --username %s
            %s
        ''', ctx.pypi.repository, ctx.pypi.username, files[0]), echo=True)


@_invoke.task()
def wheels(ctx):
    """ Upload wheels """
    with ctx.shell.root_dir():
        files = list(ctx.shell.files('wheel/dist', '*manylinux*.whl'))
        if not files:
            ctx.fail("No tarball found ")

        ctx.run(ctx.c(
            ''' twine upload --repository-url %s --username %s '''
            + ' %s ' * len(files),
            ctx.pypi.repository, ctx.pypi.username, *files
        ), echo=True)



# -*- encoding: ascii -*-
"""
invoke tasks
~~~~~~~~~~~~

"""


def namespace():
    """ Create invoke task namespace """

    class adict(object):
        """ attribute dict """
        # pylint: disable = invalid-name, missing-docstring

        def __init__(self, *args, **kwargs):
            self.__x__ = dict(*args, **kwargs)

        def __getitem__(self, name):
            return self.__x__[name]

        def __getattr__(self, name):
            if name == '__setstate__':
                raise AttributeError(name)
            try:
                return self.__x__[name]
            except KeyError:
                raise AttributeError(name)

        def items(self):
            return self.__x__.items()

    import os as _os
    import sys as _sys

    from . import _shell

    def fail(msg):
        """ Exit with message """
        _sys.stderr.write('Error: %s\n' % (msg,))
        raise _invoke.Exit(1)

    env = adict(
        package='rjsmin',
        test=adict(ignore=[]),
        doc=adict(
            userdoc="docs/userdoc",
            website=adict(
                source="docs/website",
                target="dist/website",
            ),

            sphinx=adict(
                build='docs/_userdoc/_build',
                source='docs/_userdoc',
            ),
        ),
        pypi=adict(
            # repository='https://test.pypi.org/legacy/',
            repository='https://upload.pypi.org/legacy/',
            username='ndparker',
        ),

        shell=adict((key, value) for key, value in vars(_shell).items()
                    if not key.startswith('_')),
        c=_shell.command,
        q=lambda x: _shell.command('%s', x),
        fail=fail,
    )

    _sys.path.insert(0, _os.path.dirname(
        _os.path.dirname(_os.path.abspath(__file__))
    ))

    class Vars(object):
        """ Submodules container """
        from . import (  # noqa
            benchmark,
            build,
            check,
            clean,
            compile,
            doc,
            test,
            upload,
        )

    import invoke as _invoke

    result = _invoke.Collection(*[value for key, value in vars(Vars).items()
                                  if not key.startswith('__')])
    result.configure(env)
    return result

namespace = namespace()



# -*- encoding: ascii -*-
"""
Release code
~~~~~~~~~~~~

"""

import re as _re


def check_committed(ctx):
    """ Check if everything is committed """
    git = ctx.shell.frompath("git")
    lines = ctx.run(ctx.c('%s branch --color=never', git),
                    env=dict(LC_ALL='C'), hide=True).stdout.splitlines()
    for line in lines:
        if line.startswith('*'):
            branch = line.split(None, 1)[1]
            break
    else:
        ctx.fail("Could not determine current branch.")

    if branch != 'master':
        rex = _re.compile(r'^\d+(?:\.\d+)*\.[xX]$').match
        match = rex(branch)
        if not match:
            ctx.fail("Not in master or release branch.")

    lines = ctx.run(ctx.c('%s status --porcelain', git),
                    env=dict(LC_ALL='C'), hide=True).stdout
    if lines:
        ctx.fail("Uncommitted changes!")


def add_tag(ctx):
    """ Add release tag """
    version = ctx.run('python setup.py --version', hide=True).stdout.strip()
    git = ctx.shell.frompath('git')

    ctx.run(ctx.c('''
        %s tag -a -m "Release version %s" -- %s
    ''', git, version, version), echo=True)



#!/usr/bin/env python
from __future__ import print_function

import sys

a = sys.argv[1]

while len(a) > 62:
    b, a = a[:62], a[62:]
    while b.endswith('\\'):
        b = b[:-1]
        a = '\\' + a
    print("r'%s'" % (b,))
print("r'%s'" % (a,))



# -*- encoding: ascii -*-
"""
Test suite tasks
~~~~~~~~~~~~~~~~

"""

import invoke as _invoke


@_invoke.task()
def local(ctx):
    """ Run the test suite using py.test """
    tester = ctx.shell.frompath('py.test')
    if tester is None:
        raise RuntimeError("py.test not found")

    with ctx.shell.root_dir():
        command = r'''
            %s -c test.ini -vv -s
            --doctest-modules
            --color=yes
            --exitfirst
        '''
        args = [tester]

        for ignored in ctx.test.ignore:
            command += ' --ignore %s'
            args.append(ignored)

        command += ' %s'
        args.append('tests')

        ctx.run(ctx.c(command, *args), echo=True)


@_invoke.task(default=True)
def tox(ctx, rebuild=False, env=None):
    """ Run the test suite using tox """
    tox = ctx.shell.frompath('tox')  # pylint: disable = redefined-outer-name
    if tox is None:
        raise RuntimeError("tox not found")

    command = r''' %s -c test.ini '''
    args = [tox]
    if rebuild:
        command += ' -r'
    if env is not None:
        command += ' -e %s'
        args.append(env)

    with ctx.shell.root_dir():
        ctx.run(ctx.c(command, *args), echo=True)



# -*- encoding: ascii -*-
"""
Cleanup tasks
~~~~~~~~~~~~~

"""

import invoke as _invoke


@_invoke.task()
def py(ctx):
    """ Wipe *.py[co] files """
    for name in ctx.shell.files('.', '*.py[co]'):
        ctx.shell.rm(name)
    for name in ctx.shell.dirs('.', '__pycache__'):
        ctx.shell.rm_rf(name)


@_invoke.task(py)
def dist(ctx):
    """ Wipe all """
    clean(ctx, so=True, cache=True)


@_invoke.task(py, default=True)
def clean(ctx, so=False, cache=False):
    """ Wipe *.py[co] files and test leftovers """
    for name in ctx.shell.files('.', '.coverage*', recursive=False):
        ctx.shell.rm(name)
    for name in ctx.shell.files('bench', '.out.*', recursive=False):
        ctx.shell.rm(name)
    ctx.shell.rm_rf(
        'docs/coverage',
        'docs/gcov',
        'build',
        'dist',
        'wheel/dist',
        ctx.doc.userdoc,
        'docs/_userdoc/_build',
        ctx.doc.website.source,
        ctx.doc.website.target,
    )
    if cache:
        cacheclean(ctx)
    if so:
        soclean(ctx)


@_invoke.task()
def cacheclean(ctx):
    """ Wipe Cache files """
    ctx.shell.rm_rf(
        '.tox',
        'bench/.tox',
        '.cache',
        'tests/.cache',
        'tests/.pytest_cache',
        '.mypy_cache',
    )


@_invoke.task()
def soclean(ctx):
    """ Wipe *.so files """
    for name in ctx.shell.files('.', '*.pyd'):
        ctx.shell.rm(name)
    for name in ctx.shell.files('.', '*.so'):
        ctx.shell.rm(name)



# -*- encoding: ascii -*-
"""
Checking tasks
~~~~~~~~~~~~~~

"""

import invoke as _invoke

from . import clean as _clean


@_invoke.task(_clean.py)
def lint(ctx):
    """ Run pylint """
    pylint = ctx.shell.frompath('pylint')
    if pylint is None:
        raise RuntimeError("pylint not found")

    with ctx.shell.root_dir():
        ctx.run(ctx.c(
            r''' %(pylint)s --rcfile pylintrc %(package)s ''',
            pylint=pylint,
            package=ctx.package
        ), echo=True)


@_invoke.task(_clean.py)
def flake8(ctx):
    """ Run flake8 """
    flake8 = ctx.shell.frompath('flake8')
    if flake8 is None:
        raise RuntimeError("flake8 not found")

    with ctx.shell.root_dir():
        ctx.run(ctx.c(
            r''' %(flake8)s %(package)s.py ''',
            flake8=flake8,
            package=ctx.package
        ), echo=True)


@_invoke.task(lint, flake8, default=True)
def all(ctx):
    """ Run all """



#!/usr/bin/env python
u"""
========================================
 Character table generator for rjsmin.c
========================================

Character table generator for rjsmin.c

:Copyright:

 Copyright 2011 - 2021
 Andr\xe9 Malo or his licensors, as applicable

:License:

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
"""
from __future__ import print_function
__author__ = u"Andr\xe9 Malo"
__docformat__ = "restructuredtext en"
__license__ = "Apache License, Version 2.0"

import re as _re

TPL = r"""
static const unsigned short rjsmin_charmask[128] = {
    @@mask@@
};
""".strip() + "\n"


def _make_charmask():
    """ Generate character mask table """
    # pylint: disable = too-many-branches

    dull = r'[^\047"\140/\000-\040]'
    pre_regex = r'[(,=:\[!&|?{};\r\n+*-]'
    regex_dull = r'[^/\\\[\r\n]'
    regex_cc_dull = r'[^\\\]\r\n]'

    id_literal = r'[^\000-#%-,./:-@\[-^\140{-~-]'
    id_literal_open = r'[^\000-\040"#%-\047)*,./:-@\\-^\140|-~]'
    id_literal_close = r'[^\000-!#%&(*,./:-@\[\\^{|~]'
    post_regex_off = r'[^\000-\040&)+,.:;=?\]|}-]'
    a_z = r'[a-z]'

    string_dull = r'[^\047"\140\\\r\n]'

    space = r'[\000-\011\013\014\016-\040]'

    charmask = []
    for x in range(8):  # pylint: disable = invalid-name
        maskline = []
        for y in range(16):  # pylint: disable = invalid-name
            c, mask = chr(x * 16 + y), 0
            if _re.match(dull, c):
                mask |= 1
            if _re.match(pre_regex, c):
                mask |= 2
            if _re.match(regex_dull, c):
                mask |= 4
            if _re.match(regex_cc_dull, c):
                mask |= 8
            if _re.match(id_literal, c):
                mask |= 16
            if _re.match(id_literal_open, c):
                mask |= 32
            if _re.match(id_literal_close, c):
                mask |= 64
            if _re.match(string_dull, c):
                mask |= 128
            if _re.match(space, c):
                mask |= 256
            if _re.match(post_regex_off, c):
                mask |= 512
            if _re.match(a_z, c):
                mask |= 1024

            if mask < 10:
                mask = '   ' + str(mask)
            elif mask < 100:
                mask = '  ' + str(mask)
            elif mask < 1000:
                mask = ' ' + str(mask)
            maskline.append(str(mask))
            if y == 7:
                charmask.append(', '.join(maskline))
                maskline = []
        charmask.append(', '.join(maskline))
    return TPL.replace('@@mask@@', ',\n    '.join(charmask))

print(_make_charmask())



# -*- encoding: ascii -*-
"""
Compile tasks
~~~~~~~~~~~~~

"""

import os as _os

import invoke as _invoke


@_invoke.task(default=True)
def compile(ctx):  # pylint: disable = redefined-builtin
    """ Compile the package """
    with ctx.shell.root_dir():
        ctx.run('pip install -e .',
                env=dict(_os.environ, SETUP_CEXT_REQUIRED="1"))



# -*- encoding: ascii -*-
"""
Build Tasks
~~~~~~~~~~~

"""
from __future__ import print_function

import os as _os
import platform as _platform
import re as _re

import invoke as _invoke

from . import doc as _doc
from . import _dist
from . import _release
from . import _version


@_invoke.task(_doc.doc, default=True)
def source(ctx):
    """ Build source package """
    with ctx.shell.root_dir():
        ctx.run('python setup.py sdist')


@_invoke.task()
def wheels(ctx, arches=None):
    """
    Build wheels

    Parameters:
      arches (str):
        Space separated list of architectures. If not supplied it defaults to
        the current machine type. Except if the current machine type is
        "x86_64", then if defaults to "x86_64 i686 aarch64" (using
        binfmt_misc in combination with multiarch/qemu-user-static for the
        latter).
    """
    # pylint: disable = too-many-branches

    path = 'wheel/dist'

    machine = _platform.machine()
    if arches is None:
        if machine == 'x86_64':
            arches = "x86_64 i686 aarch64"
        else:
            arches = machine
    arches = arches.split()
    for arch in arches:
        if machine == arch:
            continue
        if machine == 'x86_64' and arch == 'i686':
            continue
        ctx.run(ctx.c('''
            docker run --rm --privileged multiarch/qemu-user-static
            --reset -p yes
        '''), echo=True, pty=True)
        break

    with ctx.shell.root_dir():
        ctx.shell.rm_rf(path)
        pythons = "36 37 38 39 310"

        for arch in arches:
            if arch in ("x86_64", "i686"):
                ctx.run(
                    ctx.c('''
                        docker run --rm -it -v%s/wheel:/io
                        quay.io/pypa/manylinux1_%s:latest
                    ''' + (arch == "i686" and "linux32" or "") + '''
                        /io/build.sh %s %s
                    ''', _os.getcwd(), arch, ctx.package, "27"),
                    echo=True, pty=True,
                )
            ctx.run(
                ctx.c('''
                    docker run --rm -it -v%s/wheel:/io
                    quay.io/pypa/manylinux_2_24_%s:latest
                    ''' + (arch == "i686" and "linux32" or "") + '''
                    /io/build.sh %s %s
                ''', _os.getcwd(), arch, ctx.package, pythons),
                echo=True, pty=True,
            )

        # strip file names
        multi_sub = _re.compile(r'(?:[.-]manylinux[^.]+)+').sub
        tomove = []
        for name in _os.listdir(path):
            if not name.endswith('.whl'):
                continue
            pick = _best_manylinux(name)
            if pick:
                tomove.append((
                    name, multi_sub((lambda _, p=pick: '-' + p), name)
                ))
        for old, new in tomove:
            if old == new:
                continue

            print("%s -> %s" % (old, new))
            _os.rename(_os.path.join(path, old), _os.path.join(path, new))


def _best_manylinux(name):
    """ Find best manylinux variant produced by auditwheel """
    tag_list_search = _re.compile(r'(?:[.-]manylinux[^.]+)+').search

    tags = tag_list_search(name)
    if not tags:
        return None
    sortable = []
    for tag in tags.group(0).strip('-.').split('.'):
        assert tag.startswith('manylinux')
        ver = tag[len('manylinux'):]
        if ver.startswith(("1_", "2010_", "2014_")):
            sortable.append(((1, int(ver.split('_')[0])), tag))
            continue
        assert ver.startswith('_')
        ver = tuple(map(int, ver[1:].split('_')[:2]))
        assert ver[0] > 1
        sortable.append((ver, tag))
    if not sortable:
        return None

    return min(sortable)[-1]


@_invoke.task(_doc.doc)
def dist(ctx):
    """ Build distribution """
    fakeroot = ctx.shell.frompath('fakeroot')
    with ctx.shell.root_dir():
        ctx.shell.rm_rf('build', 'dist')
        ctx.run(ctx.c('%s python setup.py sdist --formats tar,zip', fakeroot),
                echo=True)

        files = list(ctx.shell.files('dist', '*.zip'))
        digestname = files[0][:-3] + 'digests'
        for name in ctx.shell.files('dist', '*.tar'):
            files.append(_dist.compress(ctx, name, 'gzip', '.gz'))
            files.append(_dist.compress(ctx, name, 'bzip2', '.bz2'))
            files.append(_dist.compress(ctx, name, 'xz', '.xz'))
            ctx.shell.rm(name)
        files = [name for name in files if name]

        _dist.digest(ctx, files, digestname)
        _dist.copy_changes(ctx)


@_invoke.task()
def version(ctx):
    """ Version """
    with ctx.shell.root_dir():
        _version.update(ctx)


@_invoke.task(_doc.doc)
def release(ctx):
    """ Release """
    with ctx.shell.root_dir():
        _release.check_committed(ctx)
        version(ctx)
        _release.add_tag(ctx)

    # _doc.doc(ctx)  # dependencies are not called, hence put as regular
    # dep here
    dist(ctx)



# -*- encoding: ascii -*-
"""
Benchmarks
~~~~~~~~~~

"""

import invoke as _invoke


@_invoke.task(default=True)
def tox(ctx, rebuild=False, env=None):
    """
    Run the benchmarks using tox

    :Parameters:
      `rebuild` : ``bool``
        Rebuild the tox environments?

      `env` : ``str``
        Run this one environment only.
    """
    tox = ctx.shell.frompath('tox')  # pylint: disable = redefined-outer-name
    if tox is None:
        raise RuntimeError("tox not found")

    command = r''' %s -c bench/tox.ini '''
    args = [tox]
    if rebuild:
        command += ' -r'
    if env is not None:
        command += ' -e %s'
        args.append(env)

    with ctx.shell.root_dir():
        ctx.run(ctx.c(command, *args), echo=True)



# -*- encoding: ascii -*-
"""
Release code
~~~~~~~~~~~~

"""

import errno as _errno
import os as _os


def update(ctx):
    """ Update version in relevant places """
    version = ctx.run('python setup.py --version', hide=True).stdout.strip()

    _userdoc(ctx, version)
    _download(ctx, version)
    _changes(ctx, version)


def _userdoc(ctx, version):
    """ Update version in userdoc """
    short = '.'.join(version.split('.')[:2])
    conf = _os.path.join(ctx.doc.sphinx.source, 'conf.py')

    with open(conf, 'rb') as fp:
        lines = fp.read().decode('latin-1').splitlines(True)

    seen = set()
    with open(conf, 'wb') as fp:
        for line in lines:
            if line.startswith('version'):
                line = u'version = %r\n' % (short,)
                seen.add('version')
            elif line.startswith('release'):
                line = u'release = %r\n' % (version,)
                seen.add('release')
            fp.write(line.encode('latin-1'))

    assert len(seen) == 2, "version/release not found in userdoc/conf.py"


def _download(ctx, version):  # noqa
    """ Update version in download files """
    # pylint: disable = too-many-branches, too-many-statements

    filename = _os.path.join(ctx.doc.sphinx.source, 'website_download.txt')
    isdev = 'dev' in version

    dllines, dlpath = [], ''
    if isdev:
        oldstable, hasstable = [], False
        try:
            with open(filename, 'rb') as fp:
                lines = fp.read().decode('latin-1').splitlines(True)
        except IOError as e:
            if e.errno != _errno.ENOENT:
                raise
        else:
            for line in lines:
                if line.startswith('.. begin stable'):
                    hasstable, dllines = True, oldstable
                oldstable.append(line)
        if not hasstable:
            dlpath = 'dev/'

    newdev = []
    with open(filename + '.in', 'rb') as fp:
        lines = fp.read().decode('latin-1').splitlines(True)
    if not dllines:
        dllines = lines
    else:
        for line in lines:
            if newdev:
                newdev.append(line)
                if line.startswith('.. end dev'):
                    break
            elif line.startswith('.. begin dev'):
                newdev.append(line)
        else:
            ctx.fail("Incomplete/missing dev marker in %s"
                     % (filename + '.in',))

    instable, indev = [], []
    with open(filename, 'wb') as fp:
        for line in dllines:
            if instable:
                instable.append(line)
                if line.startswith('.. end stable'):
                    if isdev:
                        res = ''.join(instable) if hasstable else ''
                    else:
                        res = (
                            ''.join(instable)
                            .replace('@@VERSION@@', version)
                            .replace('@@PATH@@', '')
                        )
                    fp.write(res.encode('latin-1'))
                    instable = []
            elif indev:
                indev.append(line)
                if line.startswith('.. end dev'):
                    if not isdev:
                        res = ''.join([indev[0], indev[-1]])
                    else:
                        res = (
                            ''.join(newdev or indev)
                            .replace('@@DEVVERSION@@', version)
                            .replace('@@PATH@@', 'dev/')
                        )
                    fp.write(res.encode('latin-1'))
                    indev = []
            elif line.startswith('.. begin stable'):
                instable.append(line)
            elif line.startswith('.. begin dev'):
                indev.append(line)
            elif isdev and hasstable:
                fp.write(line.encode('latin-1'))
            else:
                fp.write(line
                         .replace('@@VERSION@@', version)
                         .replace('@@PATH@@', dlpath)
                         .encode('latin-1'))


def _changes(ctx, version):
    """ Update version in CHANGES """
    fname = ctx.shell.native('docs/CHANGES')
    with open(fname, 'rb') as fp:
        lines = fp.read().decode('latin-1').splitlines(True)

    with open(fname, 'wb') as fp:
        for line in lines:
            if line.rstrip() == 'Changes with version':
                line = "%s %s\n" % (line.rstrip(), version)
            fp.write(line.encode('latin-1'))



# -*- encoding: ascii -*-
"""
Dist building code
~~~~~~~~~~~~~~~~~~

"""

import os as _os


def compress(ctx, filename, cmd, ext):
    """ Compress tar file """
    prog = ctx.shell.frompath(cmd)
    if prog is None:
        return None

    ctx.run('%s -c9 <%s >%s'
            % (ctx.q(prog), ctx.q(filename), ctx.q(filename + ext)),
            echo=True)
    return filename + ext


def digest(ctx, files, digestname):
    """ Make digest file """
    import hashlib as _hashlib

    result = {}
    for filename in files:
        basename = _os.path.basename(filename)
        result[basename] = []
        for method in ('md5', 'sha1', 'sha256'):
            sign = getattr(_hashlib, method)()
            with open(filename, 'rb') as fp:
                for block in iter(lambda fp=fp: fp.read(8192), b''):
                    sign.update(block)
            result[basename].append(sign.hexdigest())

    with open(digestname, 'wb') as fp:
        fp.write(b'\n# The file contains MD5, SHA1 and SHA256 digests\n')
        fp.write(b'# Check archive integrity with, e.g. md5sum -c\n')
        fp.write(b'# Check digest file integrity with PGP\n\n')

        for basename, digests in sorted(result.items()):
            for sign in digests:
                fp.write(("%s *%s\n" % (sign, basename)).encode('utf-8'))

    _sign_inline(ctx, digestname)


def _sign_inline(ctx, filename):
    """ Sign file (inline) """
    gpg = ctx.shell.frompath('gpg')
    ctx.run(ctx.c('''
        %s --output %s --clearsign -- %s
    ''', gpg, filename + '.signed', filename), echo=True)
    _os.rename(filename + '.signed', filename)


def copy_changes(ctx):
    """ Copy changelog """
    version = ctx.run('python setup.py --version', hide=True).stdout.strip()
    from_ = 'docs/CHANGES'
    to_ = 'dist/CHANGES-%s' % (version,)
    ctx.shell.cp(from_, to_)
